% This file is part of mix
% Copyright (C) 2005 Sergey Poznyakoff
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
@*A MIX TERMINAL.
This provides a terminal shell for the MIX simulator. The shell supports
following commands:


\tabskip=1em \halign{%
 \strut\sc#\hfil&\strut\sc#\hfil&\strut{\raggedright#}\cr
{\bf Synopsis}    & {\bf Minimal abbreviation}& {\bf Short description}\cr
BREAK {\it addr}  & B & Set breakpoint on address\cr
BREAK TEMP {\it addr} & BT & Set a temporary breakpoint\cr
BREAK {\it addr} {\it number} & B & Set breakpoint for {\it number} crossings\cr
CLEAR BREAK {\it addr} & CB & Clear breakpoint\cr
LIST BREAK        & LB & List breakpoints\cr
NEXT              & N & Execute next\cr
STEP              & S & Execute next and step into call\cr
ASGN {\it devnum} {\it unix-path} & A & Assign a file to \.{MIX} device\cr 
GO                & G & Run program\cr
DUMP              & DU & Dump \.{MIX} machine state\cr 
DUMP REGISTERS    & DR & Dump \.{MIX} registers\cr
DUMP MEMORY [{\it addr} [{\it addr}]]  & DM & Dump memory contents\cr
UNASM [{\it addr} [{\it addr}]] & U & Disassemble\cr           
LIST IO [{\it devnum}]  & LI & List {\sc I/O} devices\cr
CONT              & CO  & Continue after a breakpoint\cr
QUIT              & Q & Quit the terminal\cr}

@ Definitions
@s CPPFunction int
@s token c
@s line normal
@d T_EOL          '\n'
@d T_BREAK        256   
@d T_TEMP         257
@d T_BT           258
@d T_CLEAR        259
@d T_CB           260 
@d T_LIST         261 
@d T_LB           262
@d T_NEXT         263
@d T_STEP         264
@d T_ASGN         265
@d T_GO           266
@d T_DUMP         267
@d T_REGISTERS    268  
@d T_DR           269
@d T_MEMORY       270
@d T_DM           270
@d T_IO           271
@d T_LI           272
@d T_QUIT         273
@d T_CONT         274
@d T_UNASM        275
@d T_NUMBER       276
@d T_STRING       277

@d MAX_TOKENSIZE  256

@ Header files to include
@c
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <mix.h>
#include <mixsim.h>
#ifdef WITH_READLINE
# include <readline/readline.h>
#endif

@ 
@c
struct command_word {
     char *longname;
     char *shortname;
     int tok;
} cword[] = {@/
     { "BREAK",  "B",    T_BREAK     }, @/
     { "TEMP",  NULL,    T_TEMP      }, @/
     { "BT",    NULL,    T_BT        }, @/
     { "CLEAR", NULL,    T_CLEAR     }, @/
     { "CB",    NULL,    T_CB        }, @/
     { "LIST",  NULL,    T_LIST      }, @/
     { "LB",    NULL,    T_LB        }, @/
     { "NEXT",   "N",    T_NEXT      }, @/
     { "STEP",   "S",    T_STEP      }, @/
     { "ASGN",   "A",    T_ASGN      }, @/
     { "GO",     "G",    T_GO        }, @/
     { "DUMP",  "DU",    T_DUMP      }, @/
     { "REGISTERS", "REG", T_REGISTERS }, @/
     { "DR",    NULL,    T_DR        }, @/
     { "MEMORY",  "MEM",   T_MEMORY    }, @/
     { "DM",    NULL,    T_DM        }, @/
     { "IO",    NULL,    T_IO        }, @/
     { "LI",    NULL,    T_LI        }, @/
     { "CONT",  "CO",    T_CONT      }, @/
     { "UNASM",  "U",    T_UNASM     }, @/
     { "QUIT",   "Q",    T_QUIT      }, @/
     { 0 } @/
};

@
@c
struct {
     int type;                   /* Token type */
     int number;                 /* Numeric value if appropriate */
     char string[MAX_TOKENSIZE]; /* String value if appropriate */
} token;
char *buffer;                    /* Token buffer */ 
char *curp;                      /* Current position in the buffer */
char prompt[] = "MIX> ";         /* Default terminal prompt */
int interactive;                 /* T if current session is interactive,
                                    F otherwise */

@*1 Readline support.
Wherever {\mc GNU} {\tt readline} is available it is used to facilitate interaction
with the terminal. The readline mode will be automatically disabled if |interactive|
is {\mc F}.

The readline specific functions are protected by |WITH_READLINE| preprocessor variable.

@c
#ifdef WITH_READLINE

@ Readline command generator. Each invocation must return next possible expansion of
|text| or {\mc NULL} if no more expansions are available. |state| is 0 if
the function is called for the first time for given |text|.
@c
char *
mixterm_command_generator(const char *text, int state)
{
     static int i, len;
     const char *name;

     if (!state) {
	  i = 0;
	  len = strlen(text);
     }
     
     while ((name = cword[i].longname)) {
	  if (cword[i].shortname && strlen (cword[i].shortname) > strlen(name))
	       name = cword[i].shortname;
	  i++;
	  if (strncasecmp (name, text, len) == 0)
	       return strdup(name);
     }

     return NULL;
}

@ Prepare a NULL terminated array of possible command completions. |cmd| is the
command text entered so far, |start| and |end| delimit the part which should be
completed.

@c
char **
mixterm_command_completion(char *cmd, int start, int end)
{
     if (start == 0)
	  return rl_completion_matches (cmd, mixterm_command_generator);
     return NULL;
}

@ Get next available character from the |stream|
@c
static int
mixterm_getc(FILE *stream)
{
     unsigned char c;

     while (1) {
	  if (read(0, &c, 1) == 1)
	       return c;
	  if (errno == EINTR)
	       break;
     }
     return EOF;
}

@ Initialize internal readline variables.
@c
void
mixterm_readline_init()
{
     if (!interactive)
	  return;

     rl_readline_name = "mixsim";
     rl_attempted_completion_function = (CPPFunction*)mixterm_command_completion;
     rl_getc_function = mixterm_getc;
}

@ End of readline-specific section
@c
#endif

@ Read next lineful of characters. Returns pointer to allocated memory location, which
should be freed when no longer needed. 
@c
char *
mixterm_readline_internal()
{
     char *line;
     char *p;
     size_t alloclen, linelen;

     p = line = calloc (1, 255);
     if (!p)
	  abort();
     alloclen = 255;
     linelen = 0;
     for (;;) {
	  size_t n;

	  p = fgets (p, alloclen - linelen, stdin);

	  if (p)
	       n = strlen(p);
	  else {
	       free (line);
	       return NULL;
	  }

	  linelen += n;

	  /* Error.  */
	  if (linelen == 0) {
	       free (line);
	       return NULL;
	  }

	  /* Ok.  */
	  if (line[linelen - 1] == '\n') {
	       line[linelen - 1] = '\0';
	       return line;
	  } else {
	       char *tmp;
	       alloclen *= 2;
	       tmp = realloc(line, alloclen);
	       if (tmp == NULL)
		   abort();	
	       line = tmp;
	       p = line + linelen;
	  }
     }
}

@ A wrapper over |readline| call. 
@c
char *
mixterm_readline (char *prompt)
{
     if (interactive)
	  return readline(prompt);
     return mixterm_readline_internal();
}

@ When {\tt readline} is not available, provide a simplified version.
@c
#ifndef WITH_READLINE
char *
readline(char *prompt)
{
     if (prompt) {
	  printf("%s", prompt);
	  fflush(stdout);
     }

     return mixterm_readline_internal();
}

#define mixterm_readline_init()
#endif


@*1 Command line parser.

@ Copy current token into |token.string|
@c
void
copy_token()
{
    char *p;

    p = token.string;
    while (*curp && !isspace(*curp))
	*p++ = *curp++;
    *p = 0;
}

@ Advance |curp| to the next token. Copy current token into
|token.string| and store its type in |token.type|. Return token
type.
@c
int
nextkn()
{
     if (!curp) {
          char *p;

	  p = curp = readline(prompt);
	  if (curp) {
		  while (*curp && (*curp == ' ' || *curp == '\t'))
		       curp++;
	          if (*curp == 0) /* if an empty line is given, reuse the
				     latest command */
	               curp = buffer;
	          else {
	               free(buffer);
		       buffer = p;
#ifdef WITH_READLINE
		       add_history(curp);
#endif
		  }
	  }
     }
     
     if (!curp)
	  return 0;

     while (*curp && (*curp == ' ' || *curp == '\t'))
	  curp++;

     if (!*curp) {
	  curp = NULL;
	  return token.type = T_EOL;
     }
     
     if (isalpha(*curp) || *curp == '/') {
	  struct command_word *cwp;
	  
	  copy_token();
	  for (cwp = cword; cwp->longname; cwp++) {
	       if (strcasecmp(token.string, cwp->longname) == 0
		   || (cwp->shortname
		       && strcasecmp(token.string, cwp->shortname) == 0))
		    return token.type = cwp->tok;
	       
	  }
	  return token.type = T_STRING;
     }
     if (isdigit(*curp)) {
	  copy_token();
	  token.number = strtol(token.string, NULL, 0);
	  return token.type = T_NUMBER;
     }
     return token.type = *curp++;
}

@ Get MIX memory address from |token.number|. Bail out if the address
is not valid
@c
int
getaddr()
{
     int addr = token.number;
     
     if (addr < 0 || addr >= MEMSIZE) {
	  printf("LOCATION OUT OF ADDRESS SPACE\n");
	  return -1;
     } 
     return addr;
}

@ Set a breakpoint.\par
The command accepts up to three arguments. In it's simplest form it is

{\parindent = 5em\indent {\sc BREAK} {\it addr}}

where {\it addr} is a valid \.{MIX} address.

Another form

{\parindent = 5em\indent {\sc BREAK} {\it addr} {\it number}}  

sets the breakpoint which will be valid during next
{\it number} crossings.

Finally, the command

{\parindent = 5em\indent {\sc BREAK} {\sc TEMP} {\it addr}}

sets a temporary breakpoint, i.e. the one that will be removed after
it is crossed. It is equivalent to

{\parindent = 5em\indent {\sc BREAK} {\it addr} 1}

@c
void
m_break()
{
     int addr;
     int count = -1;
    
     if (nextkn() == T_TEMP) {
	  count = 1;
	  nextkn();
     }
     if (token.type != T_NUMBER) {
	  printf("USAGE: BREAK [TEMP] <ADDR> [<COUNT>]\n");
	  return;
     }
     if ((addr = getaddr()) == -1)
	  return;
     nextkn();
     if (count == -1 && token.type == T_NUMBER) {
	  count = token.number;
	  nextkn();
     } 
     if (token.type != T_EOL) {
	  printf("USAGE: BREAK [TEMP] <ADDR> [<COUNT>]\n");
	  return;
     }
     set_bp(addr, count);
}

@ A shorthand notation for BREAK TEMP.
@c
void
m_bt()
{
     int addr;
    
     if (nextkn() != T_NUMBER) {
	  printf("USAGE: BT <ADDR>\n");
	  return;
     }
     if ((addr = getaddr()) == -1)
	  return;
     set_bp(addr, 1);
}

@ Clear a breakpoint.\par
{\parindent=5em\indent{\sc CLEAR BREAK {\it addr}}}

Breakpoints are distinguished by address they are set to. The addresses can be
obtained by {\sc LIST BREAK} command.

{\it FIXME}: We should have the ability to distinguish breakpoints by their sequence
numbers, as |gdb| does.

@c
void
m_clear()
{
     int addr;
    
     if (nextkn() != T_BREAK) {
	  printf("EXPECTED BREAK\n");
	  return;
     }
     if (nextkn() != T_NUMBER
	 || (addr = getaddr()) == -1
	 || nextkn() != T_EOL) {
	  printf("USAGE: CLEAR BREAK <ADDR>\n");
	  return;
     }
     clear_bp(addr);
}

@ Shorthand notation for CLEAR BREAK.\par
{\parindent=5em\indent{\sc CB {\it addr}}}

@c
void
m_cb()
{
     int addr;
     
     if (nextkn() != T_NUMBER 
	 || (addr = getaddr()) == -1
	 || nextkn() != T_EOL) {
	  printf("USAGE: CB <ADDR>\n");
	  return;
     }
     clear_bp(addr);
}

@ List breakpoints
@c
void
m_lb()
{
     if (nextkn() != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     list_bp();
}

@ List I/O devices
@c
void
m_lio()
{
     int dev = -1;
     
     if (nextkn() == T_NUMBER) {
	  dev = token.number;
	  nextkn();
     }
     if (token.type != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     list_io(dev);
}

@ List command.\par
There are three variants of this command:

\tabskip=1em \halign{%
 \strut\sc#\hfil&\strut{\raggedright#}\cr
LIST IO & Lists all existing I/O devices\cr
LIST IO {\it num} & lists the device number {\it num}\cr
LIST BREAK &  lists all breakpoints\cr}

@c
void
m_list()
{
     switch (nextkn()) {
     case T_IO:
	  m_lio();
	  break;
     case T_BREAK:
	  m_lb();
	  break;
     default:
	  printf("NOT RECOGNIZED\n");
     }
}

@ Execute next \.{MIX} instruction
@c
void
m_next()
{
     nextkn();
     if (token.type != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     set_next(F);
     run();
}

@ Execute next \.{MIX} instruction. If it is a call, step into the called function.
@c
void
m_step()
{
     if (nextkn() != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     set_next(T);
     run();
}

@ Continue after crossing a breakpoint
@c
void
m_cont()
{
     if (nextkn() != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     run();
}

@ Assign \UNIX/ file name to \.{MIX} device number
@c
void
m_asgn()
{
     int devno;
     
     if (nextkn() != T_NUMBER) {
	  printf("ASGN <DEV> <UNIX-PATH>\n");
	  return;
     }
     devno = token.number;
     if (nextkn() == T_EOL) {
	  printf("ASGN <DEV> <UNIX-PATH>\n");
	  return;
     }
     asgn_io(devno, token.string);
}

@ Emulate \.{MIX} \.{GO} button
@c
void
m_go()
{
     if (nextkn() != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     go();
}

@ Dump registers.
@c
void
m_dr()
{
     if (nextkn() != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     dump_status();
}

@ Dump memory range.\par
{\parindent=5em\indent{\sc DM} [{\it addr} [{\it addr}]]}
Optional arguments specify start and end addresses of
the memory block to be displayed. The addresses are rounded to the nearest memory
boundary divisible by 5 (see |@<Diagnostic functions@>|.)

If no arguments are given, prints the entire \.{MIX} memory.

@c
void
m_dm()
{
     int start_loc = 0, end_loc = MEMSIZE;
     
     if (nextkn() == T_NUMBER) {
	  if ((start_loc = getaddr()) == -1) {
	       return;
	  }
	  if (nextkn() == T_NUMBER) {
	       if ((end_loc = getaddr()) == -1) 
		    return;
	       nextkn();
	  }
     }
     if (end_loc < start_loc) {
	  printf("BAD RANGE\n");
	  return;
     }
     if (token.type != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     dump_memory(start_loc, end_loc);
}

@ Dump memory or registers
@c
void
m_dump()
{
     switch (nextkn()) {
     case T_MEMORY:
	  m_dm();
	  break;
     case T_REGISTERS:
	  m_dr();
	  break;
     default:
	  dump();
	  break;
     }
}

@ Disassemble the memory range
{\parindent=5em{\sc UNASM} [{\it addr} [{\it addr}]]}
If no arguments are specified, disassembles the next instruction.
Short notation: {\sc U}

@c
void
m_unasm()
{
     extern int loc;
     int start_loc = loc, end_loc = loc+5;
     
     if (nextkn() == T_NUMBER) {
	  if ((start_loc = getaddr()) == -1) 
	       return;
	  if (nextkn() == T_NUMBER) {
	       if ((end_loc = getaddr()) == -1) 
		    return;
	       nextkn();
	  } else
	       end_loc = start_loc + 5;
     }
     if (end_loc < start_loc) {
	  printf("BAD RANGE\n");
	  return;
     }
     if (token.type != T_EOL) {
	  printf("EOL MISSING\n");
	  return;
     }
     disas(stdout, start_loc, end_loc);
}

@ Exit the terminal.
@c
void
m_quit()
{
     exit(0);
}

@ Process end of line.
Does nothing, currently;

@c
void
m_eol()
{
}

@ Indicate syntax error.
@c
void
m_err()
{
     printf("NOT RECOGNIZED\n");
     curp = NULL;
}

@ Dispatcher table for MIX terminal commands
@c
struct disp {
     int tok;            /* Token type */
     void @[@] (*fun)(void);  /* Handler function */
} main_disp[] = {
     { T_BREAK, m_break }, @/ 
     { T_BT,    m_bt    }, @/
     { T_CLEAR, m_clear }, @/
     { T_CB,    m_cb    }, @/
     { T_LIST,  m_list  }, @/
     { T_LB,    m_lb    }, @/
     { T_NEXT,  m_next  }, @/
     { T_STEP,  m_step  }, @/
     { T_ASGN,  m_asgn  }, @/
     { T_GO,    m_go    }, @/
     { T_DUMP,  m_dump  }, @/
     { T_DR,    m_dr    }, @/
     { T_DM,    m_dm    }, @/
     { T_LIST,  m_list  }, @/
     { T_LI,    m_lio   }, @/  
     { T_QUIT,  m_quit  }, @/
     { T_EOL,   m_eol   }, @/
     { T_CONT,  m_cont  }, @/
     { T_UNASM, m_unasm }, @/
     { 0 } @/
};

@ MIX terminal prompt. In homage to IBM 360 \hbox to 2em{\hfill $\smile$ }{\hskip -1em}\"
@c
char init_prompt[] = "\nMIX TERMINAL STATION READY FOR COMMUNICATION\n";

@ Main terminal function
@c
void
mixterm()
{
     struct disp *dsp;
     void @[@] (*fun)();

     interactive = isatty(0);
     mixterm_readline_init();
     printf("%s\n", init_prompt); 
     while (nextkn()) {
	  fun = m_err;
	  for (dsp = main_disp; dsp->tok; dsp++) {
	       if (token.type == dsp->tok) {
		    fun = dsp->fun;
		    break;
	       }
	  }
	  (*fun)();
     }
}


